#include "framebuffer.h"
#include "v3.h"
#include "math.h"

FrameBuffer::FrameBuffer(int u0, int v0, 
  int _w, int _h) : Fl_Gl_Window(u0, v0, _w, _h, 0) {

  w = _w;
  h = _h;
  pix = new unsigned int[w*h];
  Set(0xFFFFFFFF);

}


void FrameBuffer::draw() {

  glDrawPixels(w, h, GL_RGBA, 
    GL_UNSIGNED_BYTE, pix);

}

void FrameBuffer::Set(unsigned int color) {

  for (int i = 0; i < w*h; i++)
    pix[i] = color;

}

void FrameBuffer::Set(int u, int v, unsigned int col) {

  pix[(h-1-v)*w+u] = col;

}

void FrameBuffer::SetGuarded(int u, int v, unsigned int col) {

  if (u < 0 || u > w-1 || v < 0 || v > h-1)
    return;
  Set(u, v, col);

}

void FrameBuffer::DrawPoint(int u, int v, int psize, 
                            unsigned int color) {

  for (int vv = v-psize/2; vv <= v+psize/2; vv++) {
    for (int uu = u-psize/2; uu <= u+psize/2; uu++) {
      SetGuarded(uu, vv, color);
    }
  }

}

void FrameBuffer::DrawCircle(float uf, float vf, 
  float r, unsigned int color) {

  AABB bb;
  bb.AddPoint(V3(-r+uf, -r+vf, 0.0f));
  bb.AddPoint(V3(r+uf, r+vf, 0.0f));

  if (!ClipBB(bb))
    return;

  int lefti = (int) (bb[0][0]+0.5f);
  int righti = (int) (bb[1][0]-0.5f);
  int topi = (int) (bb[0][1]+0.5f);
  int bottomi = (int) (bb[1][1]-0.5f);

  V3 center(uf, vf, 0.0f);
  for (int v = topi; v <= bottomi; v++) {
    for (int u = lefti; u <= righti; u++) {
      V3 pixv(.5f+(float)u, .5f+(float)v, 0.0f);
      if ((center-pixv).Length() > r)
        continue;
      Set(u, v, color);
    }
  }

}

void FrameBuffer::DrawSegment(V3 p0, V3 p1, 
  unsigned int color) {

    V3 c0;
    c0.SetFromColor(color);
    DrawSegment(p0, p1, c0, c0);

}

void FrameBuffer::DrawSegment(V3 p0, V3 p1, 
  V3 c0, V3 c1) {

    float du = fabsf(p0[0]-p1[0]);
    float dv = fabsf(p0[1]-p1[1]);
    int stepsN = (du > dv) ? (int) ceilf(du) : (int) ceilf(dv);
    for (int i = 0; i < stepsN; i++) {
      float intfrac = ((float)i/(float)(stepsN-1));
      V3 p = p0 + (p1-p0)*intfrac;
      int u = (int)p[0];
      int v = (int)p[1];
      V3 c = c0 + (c1-c0)*intfrac;
      SetGuarded(u, v, c.GetColor());
    }
}

FrameBuffer* FrameBuffer::FindEdges() {

  FrameBuffer *ret;
  ret = new FrameBuffer(0, 0, w, h);
  ret->label("Edge map");

  ret->Set(0xFF000000);
  //float krn[3][3] = {
  //  {0.0f, -1.0f, 0.0f}, 
  //  {-1.0f, 4.0f, -1.0f}, 
  //  {0.0f, -1.0f, 0.0f}};
  float krn[3][3] = {
    {1.0f, 2.0f, 1.0f}, 
    {2.0f, 4.0f, 2.0f}, 
    {1.0f, 2.0f, 1.0f}};
  float sumk = 16.0f; 
  for (int v = 1; v < h-1; v++) {
    for (int u = 1; u < w-1; u++) {
      V3 currv(0.0f, 0.0f, 0.0f);
      for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
          currv = currv + Getv(u-1+j, v-1+i)*krn[i][j];
        }
      }
      //currv = V3(fabsf(currv[0]), fabsf(currv[1]), 
      //  fabsf(currv[2]));
      //currv = currv * 0.25f;
      currv = currv / sumk;
      ret->Set(u, v, currv.GetColor());
    }
  }

  return ret;

}

unsigned int FrameBuffer::Get(int u, int v) {

  return pix[(h-1-v)*w+u];

}

V3 FrameBuffer::Getv(int u, int v) {

  V3 ret;
  ret.SetFromColor(Get(u, v));
  return ret;

}

bool FrameBuffer::ClipBB(AABB &bb) {

  if (bb[0][0] > (float)w)
    return false;
  if (bb[1][0] < 0.0f)
    return false;
  if (bb[0][1] > (float)h)
    return false;
  if (bb[1][1] < 0.0f)
    return false;

  if (bb[0][0] < 0.0f)
    bb[0][0] = 0.0f;
  if (bb[0][1] < 0.0f)
    bb[0][1] = 0.0f;

  if (bb[1][0] > (float)w)
    bb[1][0] = (float)w;
  if (bb[1][1] > (float)h)
    bb[1][1] = (float)h;

  return true;

}